use std::{env, fs::File, io::stdout, path::PathBuf, process::ExitCode, str::FromStr};

use keyfork_shard::openpgp::{combine, discover_certs, openpgp::Cert, parse_messages};

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

fn validate<'a>(
    threshold: &str,
    recovery_file: &str,
    key_discovery: impl Into<Option<&'a str>>,
) -> Result<(u8, Vec<Cert>, PathBuf)> {
    let threshold = u8::from_str(threshold)?;
    let key_discovery = key_discovery.into().map(PathBuf::from);
    key_discovery.as_ref().map(std::fs::metadata).transpose()?;

    // Load certs from path
    let certs = key_discovery
        .map(discover_certs)
        .transpose()?
        .unwrap_or(vec![]);

    let recovery_file = PathBuf::from(recovery_file);
    std::fs::metadata(&recovery_file)?;

    Ok((threshold, certs, recovery_file))
}

fn run() -> Result<()> {
    let mut args = env::args();
    let program_name = args.next().expect("program name");
    let args = args.collect::<Vec<_>>();
    let (threshold, cert_list, recovery_file) = match args.as_slice() {
        [threshold, recovery_file, key_discovery] => {
            validate(threshold, recovery_file, key_discovery.as_str())?
        }
        [threshold, recovery_file] => validate(threshold, recovery_file, None)?,
        _ => panic!("Usage: {program_name} threshold recovery_file [key_discovery]"),
    };

    let mut encrypted_messages = parse_messages(File::open(recovery_file)?)?;

    let encrypted_metadata = encrypted_messages
        .pop_front()
        .expect("any pgp encrypted message");

    combine(
        threshold,
        cert_list,
        encrypted_metadata,
        encrypted_messages.into(),
        stdout(),
    )?;

    Ok(())
}

fn main() -> ExitCode {
    let result = run();
    if let Err(e) = result {
        eprintln!("Error: {e}");
        return ExitCode::FAILURE;
    }
    ExitCode::SUCCESS
}
